(*
 * Deadcode elimination.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

type env = SymbolSet.t

let env_empty =
   SymbolSet.empty

let env_add env v =
   SymbolSet.add v env

let env_union =
   SymbolSet.union

let env_equal =
   SymbolSet.equal

let env_mem env v =
   SymbolSet.mem v env

let env_remove env v =
   SymbolSet.remove v env

let env_remove_vars env vars =
   List.fold_left env_remove env vars

(************************************************************************
 * DEAD CODE ELIMINATION
 ************************************************************************)

(*
 * Remove dead expressions.
 *)
let rec dead_expr tenv venv e =
   let pos = string_pos "dead_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            dead_atom_expr tenv venv pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            dead_unop_expr tenv venv pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            dead_binop_expr tenv venv pos v ty op a1 a2 e
       | TailCall (f, args) ->
            dead_tailcall_expr tenv venv pos f args
       | LetExtCall (v, f, ty, args, e) ->
            dead_extcall_expr tenv venv pos v f ty args e
       | IntCall (ty, f, args) ->
            dead_intcall_expr tenv venv pos ty f args
       | IfThenElse (op, a1, a2, e1, e2) ->
            dead_if_expr tenv venv pos op a1 a2 e1 e2
       | SetSubscript (a1, ty, a2, a3, e) ->
            dead_set_subscript_expr tenv venv pos ty a1 a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            dead_subscript_expr tenv venv pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            dead_addr_of_subscript_expr tenv venv pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            dead_set_project_expr tenv venv pos ty a1 l a2 e
       | LetProject (v, ty, a, l, e) ->
            dead_project_expr tenv venv pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            dead_addr_of_project_expr tenv venv pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            dead_memcpy_expr tenv venv pos a1 ty a2 e
       | LetAlloc (v, ty, e) ->
            dead_alloc_expr tenv venv pos v ty e
       | LetClosure (v, ty, f, args, e) ->
            dead_closure_expr tenv venv pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            raise (IRException (pos, IRLevel 2))

(*
 * Elim this atom if the var is never used.
 *)
and dead_atom_expr tenv venv pos v ty a e =
   let pos = string_pos "dead_atom_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Elim the unop if it is never used.
 *)
and dead_unop_expr tenv venv pos v ty op a e =
   let pos = string_pos "dead_unop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Elim the binop if it is never used.
 *)
and dead_binop_expr tenv venv pos v ty op a1 a2 e =
   let pos = string_pos "dead_binop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Base case.
 * Create the environments for the free vars.
 *)
and dead_tailcall_expr tenv venv pos f args =
   let pos = string_pos "dead_tailcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * External call.
 * Can't drop the function call, because it
 * may have a side-effect on the global vars.
 *)
and dead_extcall_expr tenv venv pos v f ty args e =
   let pos = string_pos "dead_extcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Internal call.
 *)
and dead_intcall_expr tenv venv pos ty f args =
   let pos = string_pos "dead_intcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Conditional is never dropped.
 *)
and dead_if_expr tenv venv pos op a1 a2 e1 e2 =
   let pos = string_pos "dead_if_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Arrays.
 *)
and dead_subscript_expr tenv venv pos v ty a1 a2 e =
   let pos = string_pos "dead_subscript_expr" pos in
      raise (IRException (pos, NotImplemented))

and dead_set_subscript_expr tenv venv pos ty a1 a2 a3 e =
   let pos = string_pos "dead_set_subscript_expr" pos in
      raise (IRException (pos, NotImplemented))

and dead_addr_of_subscript_expr tenv venv pos v ty a1 a2 e =
   let pos = string_pos "dead_addr_of_subscript_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Structs.
 *)
and dead_project_expr tenv venv pos v ty a l e =
   let pos = string_pos "dead_project_expr" pos in
      raise (IRException (pos, NotImplemented))

and dead_set_project_expr tenv venv pos ty a1 l a2 e =
   let pos = string_pos "dead_set_project_expr" pos in
      raise (IRException (pos, NotImplemented))

and dead_addr_of_project_expr tenv venv pos v ty a l e =
   let pos = string_pos "dead_addr_of_project_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Memory operations.
 *)
and dead_memcpy_expr tenv venv pos a1 ty a2 e =
   let pos = string_pos "dead_memcpy_expr" pos in
      raise (IRException (pos, NotImplemented))

and dead_alloc_expr tenv venv pos v ty e =
   let pos = string_pos "dead_alloc_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Closure.
 *)
and dead_closure_expr tenv venv pos v ty f args e =
   let pos = string_pos "dead_closure_expr" pos in
      raise (IRException (pos, NotImplemented))

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)

(*
 * Program.
 *)
let dead_prog
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   raise (Invalid_argument "dead_prog: NotImplemented")

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
