(*
 * Function inlining and constant folding.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Fc_ir_standardize

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn pos v =
   IRException (pos, v)

let tenv_lookup = tenv_lookup raise_exn
let tenv_expand = tenv_expand raise_exn

(*
 * You will have to add a variable environment.
 *)

(*
 * Add a variable that can't be inlined.
 *)
let venv_add_type venv v ty =
   raise (Invalid_argument "venv_add_type: NotImplemented")

(************************************************************************
 * INLINING
 ************************************************************************)

(*
 * Inline an atom.
 * If the atom is a var,
 * you should look up its value from
 * the variable environment.
 *)
let inline_atom venv pos a =
   let pos = string_pos "inline_atom" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Function inlining,
 * Constant folding.
 *)
let rec inline_expr tenv venv dsym e =
   let pos = string_pos "inline_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            inline_atom_expr tenv venv dsym pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            inline_unop_expr tenv venv dsym pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            inline_binop_expr tenv venv dsym pos v ty op a1 a2 e
       | TailCall (f, args) ->
            inline_tailcall_expr tenv venv dsym pos f args
       | LetExtCall (v, f, ty, args, e) ->
            inline_extcall_expr tenv venv dsym pos v f ty args e
       | IntCall (ty, f, args) ->
            inline_intcall_expr tenv venv dsym pos ty f args
       | IfThenElse (op, a1, a2, e1, e2) ->
            inline_if_expr tenv venv dsym pos op a1 a2 e1 e2
       | SetSubscript (a1, ty, a2, a3, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let a3 = inline_atom venv pos a3 in
               SetSubscript (a1, ty, a2, a3, inline_expr tenv venv dsym e)
       | LetSubscript (v, ty, a1, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let venv = venv_add_type venv v ty in
               LetSubscript (v, ty, a1, a2, inline_expr tenv venv dsym e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let venv = venv_add_type venv v ty in
               LetAddrOfSubscript (v, ty, a1, a2, inline_expr tenv venv dsym e)
       | SetProject (a1, l, ty, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
               SetProject (a1, l, ty, a2, inline_expr tenv venv dsym e)
       | LetProject (v, ty, a, l, e) ->
            let a = inline_atom venv pos a in
            let venv = venv_add_type venv v ty in
               LetProject (v, ty, a, l, inline_expr tenv venv dsym e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            let a = inline_atom venv pos a in
            let venv = venv_add_type venv v ty in
               LetAddrOfProject (v, ty, a, l, inline_expr tenv venv dsym e)
       | Memcpy (a1, ty, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
               Memcpy (a1, ty, a2, inline_expr tenv venv dsym e)
       | LetAlloc (v, ty, e) ->
            let venv = venv_add_type venv v (TyArray (ty, None)) in
               LetAlloc (v, ty, inline_expr tenv venv dsym e)
       | LetClosure (v, ty, f, args, e) ->
            inline_closure_expr tenv venv dsym pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            let pos = string_pos "inline_expr" (atom_pos e) in
               raise (IRException (pos, IRLevel 2))

(*
 * Inline an atom.
 * This always gets inlined because it has no effect.
 *)
and inline_atom_expr tenv venv dsym pos v ty a e =
   let pos = string_pos "inline_atom_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Inline a unop.
 * Inline only if the value is a constant.
 *)
and inline_unop_expr tenv venv dsym pos v ty op a e =
   let pos = string_pos "inline_unop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Binop.
 *)
and inline_binop_expr tenv venv dsym pos v ty op a1 a2 e =
   let pos = string_pos "inline_binop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Function call.
 * Inline the function if possible.
 *)
and inline_tailcall_expr tenv venv dsym pos f args =
   let pos = string_pos "inline_tailcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Closure.
 * Save the closure in the environment,
 *)
and inline_closure_expr tenv venv dsym pos v ty f args e =
   let pos = string_pos "inline_closure_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * External call.
 * Unfold it if the call is to an internal function.
 *)
and inline_extcall_expr tenv venv dsym pos v f ty args e =
   let pos = string_pos "inline_extcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Internal call is also like a tailcall.
 *)
and inline_intcall_expr tenv venv dsym pos ty f args =
   let pos = string_pos "inline_intcall_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Conditional.
 * Remove the conditional if the test can be determined.
 *)
and inline_if_expr tenv venv dsym pos op a1 a2 e1 e2 =
   let pos = string_pos "inline_if_expr" pos in
      raise (IRException (pos, NotImplemented))

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)
(*
 * Inline a prog.
 *)
let inline_prog fun_flag
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   raise (Invalid_argument "inline_prog: NotImplemented")

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
