(*
 * Convert to continuation-passing style.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_env
open Fc_ir_exn
open Fc_ir_exn_type
open Fc_ir_standardize

(************************************************************************
 * UTILITIES
 ************************************************************************)

(*
 * Default name of the continuation argument.
 * The name is significant: we want to call
 * the "exit" function as the very last thing in
 * the program.  This symbol will get shadowed
 * inside global functions.
 *)
let cont_sym = Symbol.add "exit"

(*
 * Default exception.
 *)
let raise_exn pos v =
   IRException (pos, v)

let tenv_lookup = tenv_lookup raise_exn
let tenv_expand = tenv_expand raise_exn

(************************************************************************
 * EXPRESSIONS
 ************************************************************************)

(*
 * Continuation-passing transformation.
 *)
let rec cont_expr tenv e =
   let pos = string_pos "cont_expr" (atom_pos e) in
      match e with
         LetTypes (types, e) ->
            cont_types_expr tenv pos types e
       | LetFuns (funs, e) ->
            cont_funs_expr tenv pos funs e
       | LetAtom (v, ty, a, e) ->
            LetAtom (v, ty, a, cont_expr tenv e)
       | LetCopy (v, ty, a, e) ->
            LetCopy (v, ty, a, cont_expr tenv e)
       | LetUnop (v, ty, op, a, e) ->
            (* TODO: the type is not correct *)
            LetUnop (v, ty, op, a, cont_expr tenv e)
       | LetBinop (v, ty, op, a1, a2, e) ->
            LetBinop (v, ty, op, a1, a2, cont_expr tenv e)
       | LetApply (v, ty, f, args, e) ->
            cont_apply_expr tenv pos v ty f args e
       | TailCall (f, args) ->
            TailCall (f, args)
       | LetExtCall (v, f, ty, args, e) ->
            LetExtCall (v, f, ty, args, cont_expr tenv e)
       | Return a ->
            cont_return_expr tenv pos a
       | LetString (v, s, e) ->
            LetString (v, s, cont_expr tenv e)
       | IfThenElse (op, a1, a2, e1, e2) ->
            IfThenElse (op, a1, a2, cont_expr tenv e1, cont_expr tenv e2)
       | SetVar (v, ty, a, e) ->
            SetVar (v, ty, a, cont_expr tenv e)
       | LetAddrOfVar (v1, ty, v2, e) ->
            LetAddrOfVar (v1, ty, v2, cont_expr tenv e)
       | SetSubscript (a1, ty, a2, a3, e) ->
            SetSubscript (a1, ty, a2, a3, cont_expr tenv e)
       | LetSubscript (v, ty, a1, a2, e) ->
            LetSubscript (v, ty, a1, a2, cont_expr tenv e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            LetAddrOfSubscript (v, ty, a1, a2, cont_expr tenv e)
       | SetProject (a1, l, ty, a2, e) ->
            SetProject (a1, l, ty, a2, cont_expr tenv e)
       | LetProject (v, ty, a, l, e) ->
            LetProject (v, ty, a, l, cont_expr tenv e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            LetAddrOfProject (v, ty, a, l, cont_expr tenv e)
       | Memcpy _
       | LetAlloc _
       | LetClosure _
       | IntCall _ ->
            raise (IRException (pos, IRLevel 1))

(*
 * Add the types to the environment.
 *)
and cont_types_expr tenv pos types e =
   let pos = string_pos "cont_types_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * On global functions, add a continuation argument.
 * If a function is global, and it is defined as a LetExtCall,
 * clear the global flag.
 *)
and cont_funs_expr tenv pos funs e =
   let pos = string_pos "cont_funs_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Convert a function application to a tail call.
 *)
and cont_apply_expr tenv pos v ty f args e =
   let pos = string_pos "cont_apply_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * A return becomes a call-current-continuation.
 *)
and cont_return_expr tenv pos a =
   let pos = string_pos "cont_return_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * The CPS result may introduce duplicate variables with
 * the name "cont",. so standardize the code.
 *)
let cont_expr tenv e =
   standardize_expr (cont_expr tenv e)

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
