(*
 * Typecheck the IR.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_env

(************************************************************************
 * TYPE CHECKING
 ************************************************************************)

(*
 * Compare two types.
 *)
let rec check_types raise_exn tenv pos eq_vars ty1 ty2 =
   match ty1, ty2 with
      TyAny, _
    | _, TyAny
    | TyChar, TyChar
    | TyInt, TyInt
    | TyFloat, TyFloat ->
         ()
    | TyArray (ty1, _), TyArray (ty2, _) ->
         check_types raise_exn tenv pos eq_vars ty1 ty2
    | TyStruct fields1, TyStruct fields2 ->
         check_fields raise_exn tenv pos eq_vars fields1 fields2
    | TyFun (ty_vars1, ty_res1), TyFun (ty_vars2, ty_res2) ->
         check_types_list raise_exn tenv pos eq_vars ty_vars1 ty_vars2;
         check_types raise_exn tenv pos eq_vars ty_res1 ty_res2
    | TyId v1, TyId v2 ->
         if v1 <> v2 && not (List.mem (v1, v2) eq_vars) then
            let eq_vars = (v1, v2) :: eq_vars in
               check_types raise_exn tenv pos eq_vars (**)
                  (tenv_lookup raise_exn tenv pos v1)
                  (tenv_lookup raise_exn tenv pos v2)
    | TyId v1, _ ->
         check_types raise_exn tenv pos eq_vars (tenv_lookup raise_exn tenv pos v1) ty2
    | _, TyId v2 ->
         check_types raise_exn tenv pos eq_vars ty1 (tenv_lookup raise_exn tenv pos v2)
    | _ ->
         raise (raise_exn pos (TypeError (ty1, ty2)))

and check_types_list raise_exn tenv pos eq_vars tyl1 tyl2 =
   let len1 = List.length tyl1 in
   let len2 = List.length tyl2 in
   let _ =
      if len1 <> len2 then
         raise (raise_exn pos (ArityMismatch (len1, len2)))
   in
   let rec loop tyl1 tyl2 =
      match tyl1, tyl2 with
         ty1 :: tyl1, ty2 :: tyl2 ->
            check_types raise_exn tenv pos eq_vars ty1 ty2;
            check_types_list raise_exn tenv pos eq_vars tyl1 tyl2
       | [], [] ->
            ()
       | _ ->
            raise (Invalid_argument "Fc_ir_check.check_types_list")
   in
      loop tyl1 tyl2

and check_fields raise_exn tenv pos eq_vars fields1 fields2 =
   match fields1, fields2 with
      (v1, ty1) :: fields1, (v2, ty2) :: fields2 ->
         if v1 <> v2 then
            raise (raise_exn pos (TypeError (TyStruct fields1, TyStruct fields2)));
         check_types raise_exn tenv pos eq_vars ty1 ty2;
         check_fields raise_exn tenv pos eq_vars fields1 fields2
    | [], [] ->
         ()
    | _ ->
         raise (raise_exn pos (TypeError (TyStruct fields1, TyStruct fields2)))

(*
 * Normal versions don't have the eq_vars argument.
 *)
let check_types raise_exn tenv pos ty1 ty2 =
   check_types raise_exn tenv pos [] ty1 ty2

let check_types_list raise_exn tenv pos tyl1 tyl2 =
   check_types_list raise_exn tenv pos [] tyl1 tyl2

let check_fields raise_exn tenv pos fields1 fields2 =
   check_fields raise_exn tenv pos [] fields1 fields2

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
