(*
 * Rename all the non-global variables in a program.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

(*
 * Use symbol -> symbol map for the variables.
 *)
type env = var SymbolTable.t

(*
 * Empty env.
 *)
let env_empty = SymbolTable.create ()

(*
 * Add a new variable.
 *)
let env_add = SymbolTable.add

(*
 * Lookup a variable from the environment.
 *)
let env_lookup env v =
   try SymbolTable.find env v with
      Not_found ->
         v

(*
 * Check is a variable is defined.
 *)
let env_mem env v =
   SymbolTable.mem env v

(*
 * Special symbol that is defined while inside
 * a function body (so that we can check for nested
 * functions).
 *)
let return_sym = new_symbol_string "nested_fun"

(************************************************************************
 * STANDARDIZE
 ************************************************************************)

(*
 * Rename the vars in a type.
 *)
let rec standardize_type tenv ty =
   match ty with
      TyAny
    | TyChar
    | TyInt
    | TyFloat ->
         ty
    | TyArray (ty, i) ->
         TyArray (standardize_type tenv ty, i)
    | TyStruct fields ->
         TyStruct (List.map (fun (v, ty) -> v, standardize_type tenv ty) fields)
    | TyFun (ty_vars, ty_res) ->
         TyFun (List.map (standardize_type tenv) ty_vars, standardize_type tenv ty_res)
    | TyId v ->
         TyId (env_lookup tenv v)

(*
 * Rename the vars in an atom.
 *)
let standardize_atom tenv venv atom =
   match atom with
      AtomChar _
    | AtomInt _
    | AtomFloat _ ->
         atom
    | AtomNil ty ->
         AtomNil (standardize_type tenv ty)
    | AtomVar v ->
         AtomVar (env_lookup venv v)

let standardize_atoms tenv venv atoms =
   List.map (standardize_atom tenv venv) atoms

(*
 * Rename all the vars in the expr.
 *)
let rec standardize_expr tenv venv type_flag depth e =
   match e with
      LetTypes (types, e) ->
         standardize_types_expr tenv venv type_flag depth types e
    | LetFuns (funs, e) ->
         standardize_funs_expr tenv venv type_flag depth funs e
    | LetAtom (v, ty, a, e) ->
         standardize_atom_expr tenv venv type_flag depth v ty a e
    | LetCopy (v, ty, a, e) ->
         standardize_copy_expr tenv venv type_flag depth v ty a e
    | LetUnop (v, ty, op, a, e) ->
         standardize_unop_expr tenv venv type_flag depth v ty op a e
    | LetBinop (v, ty, op, a1, a2, e) ->
         standardize_binop_expr tenv venv type_flag depth v ty op a1 a2 e
    | LetApply (v, ty, f, args, e) ->
         standardize_apply_expr tenv venv type_flag depth v ty f args e
    | TailCall (f, args) ->
         standardize_tailcall_expr tenv venv type_flag depth f args
    | LetExtCall (v, f, ty, args, e) ->
         standardize_extcall_expr tenv venv type_flag depth v f ty args e
    | IntCall (ty, f, args) ->
         standardize_intcall_expr tenv venv type_flag depth ty f args
    | Return a ->
         standardize_return_expr tenv venv type_flag depth a
    | LetString (v, s, e) ->
         standardize_string_expr tenv venv type_flag depth v s e
    | IfThenElse (op, a1, a2, e1, e2) ->
         standardize_if_expr tenv venv type_flag depth op a1 a2 e1 e2
    | SetVar (v, ty, a, e) ->
         standardize_set_var_expr tenv venv type_flag depth v ty a e
    | LetAddrOfVar (v1, ty, v2, e) ->
         standardize_addr_of_var_expr tenv venv type_flag depth v1 ty v2 e
    | SetSubscript (a1, ty, a2, a3, e) ->
         standardize_set_subscript_expr tenv venv type_flag depth ty a1 a2 a3 e
    | LetSubscript (v, ty, a1, a2, e) ->
         standardize_subscript_expr tenv venv type_flag depth ty v a1 a2 e
    | LetAddrOfSubscript (v, ty, a1, a2, e) ->
         standardize_addr_of_subscript_expr tenv venv type_flag depth v ty a1 a2 e
    | SetProject (a1, l, ty, a2, e) ->
         standardize_set_project_expr tenv venv type_flag depth ty a1 l a2 e
    | LetProject (v, ty, a, l, e) ->
         standardize_project_expr tenv venv type_flag depth v ty a l e
    | LetAddrOfProject (v, ty, a, l, e) ->
         standardize_addr_of_project_expr tenv venv type_flag depth v ty a l e
    | Memcpy (a1, ty, a2, e) ->
         standardize_memcpy_expr tenv venv type_flag depth a1 ty a2 e
    | LetAlloc (v, ty, e) ->
         standardize_alloc_expr tenv venv type_flag depth v ty e
    | LetClosure (v, ty, f, args, e) ->
         standardize_closure_expr tenv venv type_flag depth v ty f args e

(*
 * Rename all the types.
 *)
and standardize_types tenv types =
   (* Rename all the type vars *)
   let tenv, types =
      List.fold_left (fun (tenv, types) (v, ty) ->
            let v' = new_symbol v in
            let tenv = env_add tenv v v' in
               tenv, (v', ty) :: types) (tenv, []) types
   in

   (* Rename all the types *)
   let types =
      List.map (fun (v, ty) -> v, standardize_type tenv ty) types
   in
      tenv, types

and standardize_types_expr tenv venv type_flag depth types e =
   (* Rename the types *)
   let tenv, types =
      if type_flag then
         standardize_types tenv types
      else
         tenv, types
   in

   (* Rename the body *)
   let e = standardize_expr tenv venv type_flag depth e in
      LetTypes (types, e)

(*
 * Rename functions and arguments.
 *)
and standardize_fun_expr tenv type_flag depth (venv, fundefs) ((f, gflag, f_ty, vars, body) : fundef) =
   (* Rename the function *)
   let f' = env_lookup venv f in

   (* Rename the vars *)
   let venv', vars' =
      List.fold_left (fun (venv, vars') v ->
            let v' = new_symbol v in
            let venv = env_add venv v v' in
               venv, v' :: vars') (venv, []) vars
   in
   let vars' = List.rev vars' in

   (* Rename the body *)
   let f_ty = standardize_type tenv f_ty in
   let body' = standardize_expr tenv venv' type_flag depth body in
      venv, (f', gflag, f_ty, vars', body') :: fundefs

and standardize_funs tenv venv type_flag depth funs =
   (* Rename the fun names *)
   let venv =
      List.fold_left (fun venv (f, gflag, _, _, _) ->
            let f' =
               if gflag && depth = 0 then
                  f
               else
                  new_symbol f
            in
               env_add venv f f') venv funs
   in
      (* Rename the function's vars and body *)
      List.fold_left (standardize_fun_expr tenv type_flag (succ depth)) (venv, []) funs

and standardize_funs_expr tenv venv type_flag depth funs e =
   (* Rename the funs *)
   let venv, funs = standardize_funs tenv venv type_flag depth funs in

   (* Rename the rest *)
   let e = standardize_expr tenv venv type_flag depth e in
      LetFuns (funs, e)

(*
 * Atom declaration.
 *)
and standardize_atom_expr tenv venv type_flag depth v ty a e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a = standardize_atom tenv venv a in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetAtom (v', ty, a, e)

(*
 * copy an atom.
 *)
and standardize_copy_expr tenv venv type_flag depth v ty a e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a =
      match a with
         Some a -> Some (standardize_atom tenv venv a)
       | None -> None
   in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetCopy (v', ty, a, e)

(*
 * Unary operation.
 *)
and standardize_unop_expr tenv venv type_flag depth v ty op a e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a = standardize_atom tenv venv a in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetUnop (v', ty, op, a, e)

(*
 * Binary operation.
 *)
and standardize_binop_expr tenv venv type_flag depth v ty op a1 a2 e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetBinop (v', ty, op, a1, a2, e)

(*
 * Function application.
 *)
and standardize_apply_expr tenv venv type_flag depth v ty f args e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let f = env_lookup venv f in
   let args = standardize_atoms tenv venv args in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetApply (v', ty, f, args, e)

(*
 * Tail call.
 *)
and standardize_tailcall_expr tenv venv type_flag depth f args =
   let f = env_lookup venv f in
   let args = standardize_atoms tenv venv args in
      TailCall (f, args)

(*
 * External call.
 *)
and standardize_extcall_expr tenv venv type_flag depth v f ty args e =
   let args = standardize_atoms tenv venv args in
   let ty = standardize_type tenv ty in
   let v' = new_symbol v in
   let venv = env_add venv v v' in
   let e = standardize_expr tenv venv type_flag depth e in
      LetExtCall (v', f, ty, args, e)

(*
 * Internal call.
 *)
and standardize_intcall_expr tenv venv type_flag depth ty f args =
   let args = standardize_atoms tenv venv args in
   let f = env_lookup venv f in
   let ty = standardize_type tenv ty in
      IntCall (ty, f, args)

(*
 * Return a value.
 *)
and standardize_return_expr tenv venv type_flag depth a =
   Return (standardize_atom tenv venv a)

(*
 * Allocate a string.
 *)
and standardize_string_expr tenv venv type_flag depth v s e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let e = standardize_expr tenv venv' type_flag depth e in
      LetString (v', s, e)

(*
 * Conditional expression.
 *)
and standardize_if_expr tenv venv type_flag depth op a1 a2 e1 e2 =
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e1 = standardize_expr tenv venv type_flag depth e1 in
   let e2 = standardize_expr tenv venv type_flag depth e2 in
      IfThenElse (op, a1, a2, e1, e2)

(*
 * Set a variable.
 *)
and standardize_set_var_expr tenv venv type_flag depth v ty a e =
   let v = env_lookup venv v in
   let a = standardize_atom tenv venv a in
   let e = standardize_expr tenv venv type_flag depth e in
   let ty = standardize_type tenv ty in
      SetVar (v, ty, a, e)

(*
 * Get address of a variable.
 *)
and standardize_addr_of_var_expr tenv venv type_flag depth v1 ty v2 e =
   let v1' = new_symbol v1 in
   let venv' = env_add venv v1 v1' in
   let v2 = env_lookup venv v2 in
   let venv = env_add venv v1 v1' in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetAddrOfVar (v1', ty, v2, e)

(*
 * Array operations.
 *)
and standardize_subscript_expr tenv venv type_flag depth ty v a1 a2 e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetSubscript (v', ty, a1, a2, e)

and standardize_set_subscript_expr tenv venv type_flag depth ty a1 a2 a3 e =
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let a3 = standardize_atom tenv venv a3 in
   let e = standardize_expr tenv venv type_flag depth e in
   let ty = standardize_type tenv ty in
      SetSubscript (a1, ty, a2, a3, e)

and standardize_addr_of_subscript_expr tenv venv type_flag depth v ty a1 a2 e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetAddrOfSubscript (v', ty, a1, a2, e)

(*
 * Record projection.
 *)
and standardize_project_expr tenv venv type_flag depth v ty a l e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a = standardize_atom tenv venv a in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetProject (v', ty, a, l, e)

and standardize_set_project_expr tenv venv type_flag depth ty a1 l a2 e =
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e = standardize_expr tenv venv type_flag depth e in
   let ty = standardize_type tenv ty in
      SetProject (a1, l, ty, a2, e)

and standardize_addr_of_project_expr tenv venv type_flag depth v ty a l e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let a = standardize_atom tenv venv a in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetAddrOfProject (v', ty, a, l, e)

(*
 * IR2 memory operations.
 *)
and standardize_memcpy_expr tenv venv type_flag depth a1 ty a2 e =
   let a1 = standardize_atom tenv venv a1 in
   let a2 = standardize_atom tenv venv a2 in
   let e = standardize_expr tenv venv type_flag depth e in
   let ty = standardize_type tenv ty in
      Memcpy (a1, ty, a2, e)

and standardize_alloc_expr tenv venv type_flag depth v ty e =
   let v' = new_symbol v in
   let venv' = env_add venv v v' in
   let e = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetAlloc (v', ty, e)

and standardize_closure_expr tenv venv type_flag depth v ty f args e =
   let v' = new_symbol v in
   let f' = env_lookup venv f in
   let args' = List.map (standardize_atom tenv venv) args in
   let venv' = env_add venv v v' in
   let e' = standardize_expr tenv venv' type_flag depth e in
   let ty = standardize_type tenv ty in
      LetClosure (v', ty, f', args', e')

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)

(*
 * Don't standardize the types.
 *)
let standardize_expr_notype e =
   standardize_expr env_empty env_empty false 0 e

(*
 * Global functions use a null environment.
 *)
let standardize_expr e =
   standardize_expr env_empty env_empty true 0 e

(*
 * A program.
 *)
let standardize_prog
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =

   (* Rename all types *)
   let tenv, types = standardize_types env_empty types in
   let data_type = standardize_type tenv data_type in

   (* Rename all strings *)
   let venv, strings =
      List.fold_left (fun (venv, strings) (v, s) ->
            let v' = new_symbol v in
            let venv = env_add venv v v' in
               venv, (v', s) :: strings) (env_empty, []) strings
   in

   (* Rename all funs *)
   let venv, funs = standardize_funs tenv venv true 0 funs in
   let init_sym = env_lookup venv init_sym in
      { prog_data_type = data_type;
        prog_init_sym = init_sym;
        prog_types = types;
        prog_strings = strings;
        prog_funs = funs
      }

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
